@@ -8,24 +8,7 @@ from .datatype import DataType
 
 
 logger = logging.getLogger('vaex.tasks')
-_task_register = {}
-
-
-@vaex.encoding.register("task")
-class task_encoder:
-    @staticmethod
-    def encode(encoding, task):
-        return task.encode(encoding)
-
-    @staticmethod
-    def decode(encoding, spec, df):
-        cls = _task_register[spec['task']]
-        return cls.decode(encoding, spec, df)
-
-
-def register(cls):
-    _task_register[cls.name] = cls
-    return cls
+register = vaex.encoding.make_class_registery('task')
 
 
 class Task(vaex.promise.Promise):
@@ -83,14 +66,14 @@ class Task(vaex.promise.Promise):
 # only used for testing
 @register
 class TaskSum(Task):
-    name = "sum-test"
+    snake_name = "sum-test"
 
     # def __init__(self, df, expression):
     #     super().__init__(df, expression)
     #     self.expression = expression
 
     def encode(self, encoding):
-        return {'task': type(self).name, 'expression': self.expressions}
+        return {'expression': self.expressions}
 
     @classmethod
     def decode(cls, encoding, spec, df):
@@ -100,12 +83,12 @@ class TaskSum(Task):
 @register
 class TaskFilterFill(Task):
     cacheable = False
-    name = "filter_fill"
+    snake_name = "filter_fill"
     def __init__(self, df):
-        super().__init__(df=df, pre_filter=True, name=self.name)
+        super().__init__(df=df, pre_filter=True, name=self.snake_name)
 
     def encode(self, encoding):
-        return {'task': type(self).name}
+        return {}
 
     @classmethod
     def decode(cls, encoding, spec, df):
@@ -113,9 +96,9 @@ class TaskFilterFill(Task):
 
 @register
 class TaskSetCreate(Task):
-    name = "set_create"
+    snake_name = "set_create"
     def __init__(self, df, expression, flatten, unique_limit=None, selection=None):
-        super().__init__(df=df, expressions=[expression], pre_filter=df.filtered, name=self.name)
+        super().__init__(df=df, expressions=[expression], pre_filter=df.filtered, name=self.snake_name)
         self.flatten = flatten
         self.dtype = self.df.data_type(expression)
         self.dtype_item = self.df.data_type(expression, axis=-1 if flatten else 0)
@@ -123,7 +106,7 @@ class TaskSetCreate(Task):
         self.selection = selection
 
     def encode(self, encoding):
-        return {'task': type(self).name, 'expression': self.expressions[0], 'dtype': encoding.encode('dtype', self.dtype),
+        return {'expression': self.expressions[0], 'dtype': encoding.encode('dtype', self.dtype),
                 'dtype_item': encoding.encode('dtype', self.dtype_item), 'flatten': self.flatten, 'unique_limit': self.unique_limit,
                 'selection': self.selection}
 
@@ -133,7 +116,7 @@ class TaskSetCreate(Task):
 
 @register
 class TaskMapReduce(Task):
-    name = "map_reduce"
+    snake_name = "map_reduce"
     cacheable = False  # in general this is used with side effects
 
     def __init__(self, df, expressions, map, reduce, info=False, to_float=False,
@@ -151,7 +134,7 @@ class TaskMapReduce(Task):
         self.selection = selection
 
     def encode(self, encoding):
-        return {'task': type(self).name, 'expressions': self.expressions, 'map': self._map, 'reduce': self._reduce,
+        return {'expressions': self.expressions, 'map': self._map, 'reduce': self._reduce,
                 'info': self.info, 'to_float': self.to_float, 'to_numpy': self.to_numpy,  # 'ordered_reduce': self.ordered_reduce,
                 'skip_masked': self.skip_masked, 'ignore_filter': self.ignore_filter, 'selection': self.selection, 'pre_filter': self.pre_filter,
                 }
@@ -276,10 +259,10 @@ OP_FIRST = StatOpFirst(6)
 
 @register
 class TaskStatistic(Task):
-    name = "legacy_statistic"
+    snake_name = "legacy_statistic"
 
     def encode(self, encoding):
-        return {'task': type(self).name, 'expressions': self.expressions,
+        return {'expressions': self.expressions,
                 'shape': self.shape, 'selections': self.selections, 'op': encoding.encode('_op', self.op), 'weights': self.weights,
                 'dtype': encoding.encode('dtype', DataType(self.dtype)), 'minima': self.minima, 'maxima': self.maxima, 'edges': self.edges,
                 'selection_waslist': self.selection_waslist}
@@ -287,7 +270,6 @@ class TaskStatistic(Task):
     @classmethod
     def decode(cls, encoding, spec, df):
         spec = spec.copy()
-        del spec['task']
         spec['op'] = encoding.decode('_op', spec['op'])
         spec['dtype'] = encoding.decode('dtype', spec['dtype'])
         selection_waslist = spec.pop('selection_waslist')
@@ -343,28 +325,35 @@ class TaskStatistic(Task):
 @register
 class TaskAggregations(Task):
     """Multiple aggregations on a single grid."""
-    name = "aggregations"
+    snake_name = "aggregations"
 
-    def __init__(self, df, grid):
-        expressions = [binner.expression for binner in grid.binners]
+    def __init__(self, df, binners):
+        expressions = [binner.expression for binner in binners]
         self.df = df
-        self.parent_grid = grid
+        self.binners = binners
         self.aggregation_descriptions = []
         self.dtypes = {}
         Task.__init__(self, df, expressions, name="statisticNd", pre_filter=df.filtered)
+        self.original_tasks = []
+
+    def __repr__(self):
+        encoding = vaex.encoding.Encoding()
+        state = self.encode(encoding)
+        import yaml
+        return yaml.dump(state, sort_keys=False, indent=4)
 
     def encode(self, encoding):
         # TODO: get rid of dtypes
-        return {'task': type(self).name,
-                'grid': encoding.encode('grid', self.parent_grid),
+        return {
+                'binners': encoding.encode_list('binner', self.binners),
                 'aggregations': encoding.encode_list("aggregation", self.aggregation_descriptions),
                 'dtypes': encoding.encode_dict("dtype", self.dtypes)
                 }
 
     @classmethod
     def decode(cls, encoding, spec, df):
-        grid = encoding.decode('grid', spec['grid'])
-        task = cls(df, grid)
+        binners = encoding.decode_list('binner', spec['binners'])
+        task = cls(df, binners)
         aggs = encoding.decode_list('aggregation', spec['aggregations'])
         for agg in aggs:
             agg._prepare_types(df)
@@ -372,6 +361,7 @@ class TaskAggregations(Task):
         return task
 
     def add_aggregation_operation(self, aggregator_descriptor):
+        assert self._fingerprint is None, "Adding operation after fingerprint is calculated"
         task = Task(self.df, [], "--")
 
         def chain_reject(x):
@@ -393,3 +383,40 @@ class TaskAggregations(Task):
     def check(self):
         if not self.aggregation_descriptions:
             raise RuntimeError('Aggregation tasks started but nothing to do, maybe adding operations failed?')
+
+
+@register
+class TaskAggregation(Task):
+    """Single aggregation on a single grid."""
+    snake_name = "aggregation-single"
+
+    def __init__(self, df, binners, aggregation_description):
+        expressions = [binner.expression for binner in binners]
+        self.df = df
+        self.binners = binners
+        self.aggregation_description = aggregation_description
+        self.dtypes = {}
+        Task.__init__(self, df, expressions, name=self.snake_name, pre_filter=df.filtered)
+        self.dtypes = {expr: self.df.data_type(expr).index_type for expr in self.expressions_all}
+        self.expressions_all.extend(aggregation_description.expressions)
+
+    def __repr__(self):
+        encoding = vaex.encoding.Encoding()
+        state = self.encode(encoding)
+        import yaml
+        return yaml.dump(state, sort_keys=False, indent=4)
+
+    def encode(self, encoding):
+        # TODO: get rid of dtypes
+        return {
+                'binners': encoding.encode_list('binner', self.binners),
+                'aggregation': encoding.encode("aggregation", self.aggregation_description),
+                'dtypes': encoding.encode_dict("dtype", self.dtypes)
+                }
+
+    @classmethod
+    def decode(cls, encoding, spec, df):
+        binners = tuple(encoding.decode_list('binner', spec['binners']))
+        agg = encoding.decode('aggregation', spec['aggregation'])
+        task = cls(df, binners, agg)
+        return task
